'''
Created on May 20, 2013

@author: mwallers
'''

"""
input: fixed (desktop) urls; fileLocation of where the content folder is located
output: fixed urls of which a mobile version was not found

along the way it records to file if the urls were found or not to the respective files


useragent.py attempts to find a mobile version of a given URL by requesting with
different user agents.

Notes:
	* Only finds mobile pages given automatically through user agents.
	* Does not work properly with window.location=[url] redirects
	  (this is a PhantomJS issue). Often it detects the presence of a
	  mobile page, but not the actual URL.

How it works:
	1. Request the URL with fixed and mobile user agents.
	2. If the mobile request ends up at a different URL than the fixed
	   request, assume the mobile URL is a mobile version of the page.
	3. Calculate the tag frequency distribution analysis (TFDA) score
	   for the fixed and mobile requests (see tfda.py).
	4. If the score is over 0 (indicating a difference), request copies
	   of the fixed version and calculate the mean and standard
	   deviation of TFDA scores between the copies and the original.
	5. If the TFDA score of the mobile request is two standard
	   deviations above the mean score of the fixed request copies,
	   consider the difference significant. This means the URL itself
	   has a mobile version (instead of providing a redirect).
	   
useragent.py will print the output for the Google version if it is turned on (via isGoogleVersion) at the same time as printing the original version's output.
This will keep the program from looping through everything twice and it will save time in the end for this to take on the switching between versions itself.
"""
 
import config
from ..util.request import requestPage
from ..util.logger import log
from ..util.confirmURLDifference import differenceExists, get_last_score
from ..util import record
import os
import sys
from mediaQueries import findMediaQueriesInStylesheets

from differingURLs import lookForAlternateLinksForGoogleVersion, lookForCanonicalLinksForGoogleVersion

			
def checkAlternateAndCanonicalLinks(fixedURL, mobileURL, fixedContent, mobileContent):
	mobileLinks = []
	mobileLinks = lookForAlternateLinksForGoogleVersion(fixedURL, fixedContent)
	if not mobileLinks: #if empty list
		log("\t->-> missing use of alternate links")
	elif not (fixedURL in mobileLinks):
		log("\t->-> improper use of alternate links")
	else:
		log("\t->-> correct use of alternate links")
		
	fixedLinks = []
	fixedLinks = lookForCanonicalLinksForGoogleVersion(mobileURL, mobileContent)
	if not fixedLinks: #if empty list
		log("\t->-> missing use of canonical links")
	elif not (fixedURL in fixedLinks):
		log("\t->-> improper use of canonical links")
	else:
		log("\t-> correct use of canonical links")


def getMobileVersion(urls, fileLocation, fixedDictionary, mobileDictionary):
	remainingURLs = []
	for url in urls:
		'''commented out section is below:'''
		
		followsRulesFromGoogle = True
		
		log('useragent find %s' % url)
		
		#make sure that the url starts with http:// or https://
		position = url.find('://')
		if position == -1:
			newURL = 'http://' + url  #add http:// as the beginning
			url = newURL
		isError = False
		fixedJs = []
		fixedContent = None
		fixedStylesheets = []
		headers = []
		fixedActuallySuccess = True
		fixedURL = fixedDictionary[url]
		
		try:
			filename = config.getDomain(url, False, True) #********************************************************************
			filename = filename + "-fixed"
			filename = fileLocation + filename
			#print "filename trying to open is: " + filename
			
			filename2 = filename  + '.html'
			htmlfile = open(filename2, 'r')
			fixedContent = htmlfile.read()
			htmlfile.close()
			
			filename2 = filename + '.css'
			ssfile = open(filename2, 'r')
			ssblock = ssfile.read()
			fixedStylesheets = ssblock.split(',')
			ssfile.close()
			
			filename2 = filename  + '.js'
			jsfile = open(filename2, 'r')
			jsblock = jsfile.read()
			fixedJs = jsblock.split(',')
			jsfile.close()
			
			filename2 = filename + '.headers'
			headersfile = open(filename2, 'r')
			headers = headersfile.readlines()
			headersfile.close()     

		except IOError as e:
			isError = True
			log("error opening the fixed files...")
			print ("failed file name is: " , filename2)
			print "I/O error({0}): {1}".format(e.errno, e.strerror)
			#if the file we broke on was NOT the html file
			fixedActuallySuccess = False
			if(filename2 != filename + '.html'): 
				isError = False
				fixedActuallySuccess = True
				print "actually a success"
		except:
			isError = True
			print "unexpected error"
			log ("\tunexpected error!", False)
			fixedActuallySuccess = False

		if isError:
			record.record_notfound(url, config.neverFoundFilename, None, None, True)
			

		#look through the header and determine if there is the vary HTTP header
		#  should say: 'Vary: User-Agent'
		#NOTE: imdb.com has some user-agent vary headers. however, they were 
		#    formatted either like: 'Vary: User-agent' OR like: 'Vary; Accept-Encoding, User-Agent'
		#    thus be sure that we look for vary header, then on the line look for user-agent. 
		varyHttpExists = False
		for header in headers:    
			if 'Vary: ' in header and 'User-Agent' in header:
				varyHttpExists = True
			if varyHttpExists:
				break                
		        
		log('mobile...')
		mobileContent = None
		mobileJs = []
		mobileStylesheets = []
		mobileHeaders = []
		
		mobileURL = mobileDictionary[url]
		
		needRepeat = True
		isFirstAttempt = True
		isMobileError = False
			
		filename = config.getDomain(url, False, True)  
		filename = filename + "-mobile"
		filename = fileLocation + filename
		try:
			filename2 = filename  + '.html'
			htmlfile = open(filename2, 'r')
			#print "\topening " + filename2
			mobileContent = htmlfile.read()
			htmlfile.close()
			
			filename2 = filename + '.css'
			ssfile = open(filename2, 'r')
			#print "\topening " + filename2
			ssblock = ssfile.read()
			mobileStylesheets = ssblock.split(',')
			ssfile.close()
			
			filename2 = filename  + '.js'
			jsfile = open(filename2, 'r')
			#print "\topening " + filename2
			jsblock = jsfile.read()
			mobileStylesheets = jsblock.split(',')
			jsfile.close()
			
			filename2 = filename + '.headers'
			headersfile = open(filename2, 'r')
			#print "\topening " + filename2
			mobileHeaders = headersfile.readlines()
			headersfile.close()   
		except IOError as e:
			isMobileError = True
			log("\terror opening the mobile files...")
			print ("failed file name is: " , filename2)
			print "I/O error({0}): {1}".format(e.errno, e.strerror)
		except:
			isMobileError = True
			print "unexpected error"
			log ("\tunexpected error!")
		

		isFlaggedMatchMedia = False
		isFlaggedUserAgent = False
		#look through the content and see if it is attempting to use client-side redirect
		if mobileContent:
			lines = mobileContent.split('\n')
			for line in lines:
				if line.find("navigator.userAgent") >= 0 and not isFlaggedUserAgent:
					#flag it as trying to use client-side reidrect
					log("\t\tFLAG: attempting to redirect on client-side")
					isFlaggedUserAgent = True
					
				if "matchMedia" in line:
					if line.find(r"matchMedia(.*)\.matches") >=0 and not isFlaggedMatchMedia:
						log("\t\tFLAG2: attempting to use javascript's matchMedia()")
						log("\ton line:" + line)
						isFlaggedMatchMedia = True #break
				if isFlaggedUserAgent and isFlaggedMatchMedia:
					break
					
		if not isFlaggedMatchMedia and mobileJs:
			for js in mobileJs:
				#get the file
				
				try:
					#Need to request with a common UA,
					#else some sites won't give it to you
					r=urllib2.Request(stylesheet)
					mobileUserAgent = MOBILE_UA
					r.add_header('User-Agent', mobileUserAgent)
					
					data=urllib2.build_opener().open(r).read()
					
				except Exception as e:
					#If the stylesheet request fails
					continue    #to the next ss
				
				for line in data:
					if line.find(r"matchMedia(.*)\.matches") >= 0:
						log("\t\tFLAG2: using javascript's matchMedia()")
						log("\ton line:" + line)
						isFlaggedMatchMedia = True 
						break
		
		'''If (isGoogleVersion and "vary HTTP: user agent" exists) or (!isGoogleVersion)
		then see if there is a significant difference in the urls or css or tfda for the 2 different urls
		'''
		if (config.isGoogleVersion and not varyHttpExists):   #isGoogleVersion and doesn't say vary http
			followsRulesFromGoogle = False
			record.printGoogleDifferent = True
			record.record_notfound(url, config.notFoundFilenameGoogleVersion, config.statsFilenameGoogleVersion, "no vary HTTP header")
			record.printGoogleDifferent = False
			log("\t-> not found because no vary HTTP header")        
		
		canBeARedirect = True
		if (not fixedURL) or fixedURL == "" or fixedURL == " ":
			fixedURL = url
			canBeARedirect = False 
			#otherwise it will mess up some websites that are actually not redirecting and dont 
			#	have a mobile page at all. if it is a true redirect then the page will look  
			#	different or have a different url and thus it will be found out later.
		if (not mobileURL) or mobileURL == "" or mobileURL == " ":
			mobileURL = fixedURL
			#canBeARedirect = False
			
		#If we've gotten errors in the request, skip the URL
		#if not fixedStatus==mobileStatus=='success':
		if not isError and isMobileError:  #means the mobile failed but fixed didnt
			#log('\terror: fixed %s, mobile %s' % (fixedStatus, mobileStatus))
			if config.isGoogleVersion:
				record.record_error(url, fixedURL, "", mobileURL, "", config.errorFileNameGoogleVersion)
			record.record_error(url, fixedURL, "", mobileURL, "", config.errorFilename)
			if fixedActuallySuccess: #only skip if just the mobile page failed and not the desktop page
				remainingURLs.append(url)
			else:
				record.record_notfound(url, config.neverFoundFilename, None, None, True)
			continue 
		
		mobileFound = False
		#If the final URLs are different, we've probably been
		#redirected to a mobile site
		#if fixedURL in mobileURL:
		newFixedURL = fixedURL + '/'
		if newFixedURL == mobileURL:
			print "\t*fixed url inside mobile url"
			log("\t*fixed url inside mobile url")
			canBeARedirect = False
		secureFixedURL  = fixedURL
		secureFixedURL = secureFixedURL.replace('http://', 'https://')
		secureMobileURL  = mobileURL
		secureMobileURL = secureMobileURL.replace('http://', 'https://')
		if secureFixedURL == secureMobileURL:
			print "\t*secure http almost marked as redirect"
			log("\t*secure http almost marked as redirect")
			canBeARedirect = False
		#print secureFixedURL + " vs " + secureMobileURL + '\n'		
		secureFixedURL += "/"
		if secureFixedURL == mobileURL or secureFixedURL == secureMobileURL:
			print "\t*secure fixed url inside mobile url"
			log("\t*secure fixed url inside mobile url")
			canBeARedirect = False
		
		if fixedURL!=mobileURL and canBeARedirect:
			log("\tjust a redirect...")
			print "fixed: "+ fixedURL + "\tmobile: " +mobileURL
			record.record_found(fixedURL, mobileURL, config.foundFilename, config.statsFilename, 'redirect')
			if config.isGoogleVersion and followsRulesFromGoogle:
				record.printGoogleDifferent = True
				record.record_found(fixedURL, mobileURL, config.foundFilenameGoogleVersion, config.statsFilenameGoogleVersion, 'redirect')
				record.printGoogleDifferent = False
			mobileFound = True
			#check for alternate and canonical links:
			checkAlternateAndCanonicalLinks(fixedURL, mobileURL, fixedContent, mobileContent)
			#cleanup directories to make room:
			print "deleting "+url+"...\n"
			config.cleanupUnusedDirectories (url);

		#what about the actual content that was returned
		if not mobileFound:
			#Confirm there's a significant difference between the versions   #the mobile error is to make sure that the mobile files passed are vaid
			try:
				if differenceExists(fixedURL, fixedContent, fixedStylesheets, fixedJs, mobileURL, mobileContent, mobileStylesheets, mobileJs, fileLocation): #NOTE: added location            
					if config.isGoogleVersion and followsRulesFromGoogle:
						record.printGoogleDifferent  = True
						record.record_found(fixedURL, mobileURL, config.foundFilenameGoogleVersion, config.statsFilenameGoogleVersion, get_last_score())
						record.printGoogleDifferent = False
					record.record_found(fixedURL, mobileURL, config.foundFilename, config.statsFilename, get_last_score())
					#check for alternate and canonical links:
					#checkAlternateAndCanonicalLinks(fixedURL, mobileURL, fixedContent, mobileContent)
					mobileFound = True
					#cleanup directories to make room:
					print "deleting "+url+"...\n"
					#config.cleanupUnusedDirectories (url);
			except:
				print "ERROR in differenceExists"
			
		#is there a difference in the stylesheets of the 2 pages?
		if not mobileFound:
			compareSS = True
			#If stylesheets are different, they probably have mobile CSS
			#if empty stylesheets for both, dont do it. if empty stylesheets for just mobile, check for a valid image
			if not mobileStylesheets or mobileStylesheets[0] == '':
				if not fixedStylesheets:
					compareSS = False
				else:
					#TODO: call image validator. then determine if the comparison should stop or not
					#&&&&&&&**************************&&&&&@@@@@@@@@@@@@@@@@@&&&&&&&!!!!!!!!!!
					compareSS = False
			if compareSS:
				fixed_ss_set=set(fixedStylesheets)
				mobile_ss_set=set(mobileStylesheets)
				ss_diff=fixed_ss_set.difference(mobile_ss_set)
				ss_diff.update(mobile_ss_set.difference(fixed_ss_set))
				if ss_diff:
					log('\tstylesheets differing so now looking for media queries in case they are connected...')
					#call mediaQueries function to determine if the ss are different because of mediaQueries!!!
					sheets = set()
					
					#add the fixed and mobile ss together to look through them all for the media queries  
					sheets.update(mobile_ss_set)
					sheets.update(fixed_ss_set)
					foundMedia, foundMediaWidth = findMediaQueriesInStylesheets(sheets, url)
					#if both a ss and mediaquery difference:
					if foundMedia and foundMediaWidth:
						log('\tstylesheets difference and media width')
						record.record_found(fixedURL, mobileURL,config.foundFilename, config.statsFilename, 'stylesheets and media width')
						#cleanup directories to make room:
						print "deleting "+url+"...\n"
						config.cleanupUnusedDirectories (url);
					else:
						#if just a ss difference:
						log('\tstylesheets difference')
						record.record_found(fixedURL, mobileURL,config.foundFilename, config.statsFilename, 'stylesheets')
						#cleanup directories to make room:
						print "deleting "+url+"...\n"
						config.cleanupUnusedDirectories (url);
					
					if config.isGoogleVersion and followsRulesFromGoogle:
						record.printGoogleDifferent = True
						if foundMedia and foundMediaWidth:
							record.record_found(fixedURL, mobileURL, config.foundFilenameGoogleVersion, config.statsFilenameGoogleVersion, 'stylesheets and media width')
							#cleanup directories to make room:
							print "deleting "+url+"...\n"
							config.cleanupUnusedDirectories (url);
						else:
							record.record_found(fixedURL, mobileURL, config.foundFilenameGoogleVersion, config.statsFilenameGoogleVersion, 'stylesheets')
							#cleanup directories to make room:
							print "deleting "+url+"...\n"
							config.cleanupUnusedDirectories (url);
						record.printGoogleDifferent = False
					mobileFound = True
					continue
		
		if not mobileFound:
			mobileURL = "none"
			#log("\tnot found by any method")
			if config.isGoogleVersion and followsRulesFromGoogle:
				record.printGoogleDifferent = True
				record.record_notfound(url, config.notFoundFilenameGoogleVersion, config.statsFilenameGoogleVersion, get_last_score())
				record.printGoogleDifferent = False
			record.record_notfound(url, config.notFoundFilename, config.statsFilename, get_last_score())
				
		'''
			#what about differing js?
			if not mobileFound:
				fixed_js_set=set(fixedJs)
				mobile_js_set=set(mobileJs)
				js_diff=fixed_js_set.difference(mobile_js_set)
				js_diff.update(mobile_js_set.difference(fixed_js_set))
				if js_diff:
					log('\tjavascript difference')
					record_found(fixedURL, mobileURL,config.foundFilename, config.statsFilename, 'javascript')
					if config.isGoogleVersion and followsRulesFromGoogle:
						record_found(fixedURL, mobileURL, config.foundFilenameGoogleVersion, config.statsFilenameGoogleVersion, 'javascript')
					mobileFound = True
		'''
		
		if mobileURL == "none":
			remainingURLs.append(url)
		
		print "made it to here in userAgent with url: " + url + "\n"
	return remainingURLs
	#return urls

if __name__ == '__main__':
	urls = []
	urls.append('http://www.cnn.com')
	getMobileVersion(urls, '/secretContent')